@@ -1,9 +1,9 @@
-/* Copyright 1999-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -140,11 +140,10 @@ static void *overlay_extension_mappings(apr_pool_t *p,
                                         const void *base_val,
                                         const void *data)
 {
-    extension_info *new_info = apr_palloc(p, sizeof(extension_info));
     const extension_info *overlay_info = (const extension_info *)overlay_val;
     const extension_info *base_info = (const extension_info *)base_val;
+    extension_info *new_info = apr_pmemdup(p, base_info, sizeof(extension_info));
 
-    memcpy(new_info, base_info, sizeof(extension_info));
     if (overlay_info->forced_type) {
         new_info->forced_type = overlay_info->forced_type;
     }
